# Holds various implementations of a "fetcher", which is given a display string
# to know its location.
# during operation, it is given a row index, and responds with the data.

from amaranth import Module, Signal, unsigned
from amaranth.build import Platform
from amaranth.lib import wiring, data
from amaranth.lib.wiring import In, Out
from amaranth.lib import stream
import logging
from itertools import pairwise

from .common import Rgb888Layout
from .geom import DisplayString


logger = logging.getLogger(__name__)

# FIXME: sizing should be based off of screen size.
CoordLayout = data.StructLayout({"x": unsigned(10), "y": unsigned(10)})


class AddressConverter(wiring.Component):
    """Translates display (x,y) into full screen (x,y) based on geometry"""

    def __init__(self, geom: DisplayString, *, src_loc_at=0):
        self.geom = geom
        super().__init__(
            {
                "input_x": In(geom.dimensions.length),
                "addr": In(unsigned(geom.dimensions.addr_bits)),
                "output": Out(data.ArrayLayout(CoordLayout, geom.dimensions.mux)),
            },
            src_loc_at=src_loc_at,
        )

    def elaborate(self, platform: Platform) -> Module:
        m = Module()

        for mux in range(self.geom.dimensions.mux):
            o = self.geom.translate_coord(self.input_x, self.addr, mux)
            m.d.comb += self.output[mux]["x"].eq(o["x"])
            m.d.comb += self.output[mux]["y"].eq(o["y"])

        return m


class AddressGenerator(wiring.Component):
    """Generates (x,y) sequences corresponding to a display row."""

    def __init__(self, geom: DisplayString, *, src_loc_at=0):
        self.geom = geom
        super().__init__(
            {
                "output": Out(
                    stream.Signature(data.ArrayLayout(CoordLayout, geom.dimensions.mux))
                ),
                "start": In(1),
                "done": Out(1),
                "addr": In(geom.dimensions.addr_bits),
            },
            src_loc_at=src_loc_at,
        )

    def elaborate(self, platform: Platform) -> Module:
        m = Module()

        counter = Signal(self.geom.dimensions.length)

        addr = Signal(self.addr.shape())

        # based on the geometry we generate x,y pairs.
        m.submodules.translate = translate = AddressConverter(self.geom)
        m.d.comb += translate.input_x.eq(counter)
        m.d.comb += translate.addr.eq(addr)

        m.d.comb += self.output.payload.eq(translate.output)

        with m.FSM():
            with m.State("init"):
                m.d.comb += [self.done.eq(0), self.output.valid.eq(0)]
                m.d.sync += [counter.eq(0), addr.eq(self.addr)]
                with m.If(self.start):
                    m.next = "run"

            with m.State("run"):
                m.d.comb += self.output.valid.eq(1)
                # stream data out as long as it's valid.

                with m.If(
                    self.output.ready & (counter == self.geom.dimensions.length - 1)
                ):
                    m.next = "done"
                with m.Elif(self.output.ready):
                    m.d.sync += counter.eq(counter + 1)

                pass
            with m.State("done"):
                m.d.comb += self.output.valid.eq(0)
                m.d.comb += self.done.eq(1)
                m.next = "init"

        return m


def example_rgb_transform(x, y):
    """A simple coordinate-RGB transformation that computes RGB values directly from the x-y position
    of the pixel. This is used in the simple case and as a test for the rest of the system."""
    return {
        "red": x + y,
        "green": x - y,
        "blue": x ^ y,
    }


class BasicFetcher(wiring.Component):
    """A generic function-based fetcher. Takes a function of the form f(x,y: int) -> dict rgb values.

    If no function is provided it uses a basic coordinate-driven rgb transform where red = x+y,
    green = x - y, and blue = x ^ y.

    When providing a function, it must return a dictionary with the keys "red", "green", "blue"."""

    def __init__(
        self,
        geom: DisplayString,
        dfunc=example_rgb_transform,
        data_shape=Rgb888Layout,
        *,
        src_loc_at=0,
    ):
        self.geom = geom
        self.dfunc = dfunc
        super().__init__(
            {
                "input": In(
                    stream.Signature(data.ArrayLayout(CoordLayout, geom.dimensions.mux))
                ),
                "output": Out(
                    stream.Signature(data.ArrayLayout(data_shape, geom.dimensions.mux))
                ),
            },
            src_loc_at=src_loc_at,
        )

    def elaborate(self, platform: Platform) -> Module:
        m = Module()

        colors = self.output.payload
        m.d.comb += [
            self.output.valid.eq(self.input.valid),
            self.input.ready.eq(self.output.ready),
        ]

        for i in range(self.geom.dimensions.mux):
            inp = self.input.payload[i]
            output = self.dfunc(inp.x, inp.y)
            m.d.comb += [
                colors[i].red.eq(output["red"]),
                colors[i].green.eq(output["green"]),
                colors[i].blue.eq(output["blue"]),
            ]

        return m


def chain_streams(m, streams):
    """For "stream combinators", this allows you to easily chain outputs to inputs."""
    for pair in pairwise(streams):
        wiring.connect(m, pair[0].output, pair[1].input)
