from amaranth import Module, Cat, Mux, ShapeLike, Signal, Assert, Array
from amaranth.build import Platform
from amaranth.lib import wiring, data
from amaranth.lib.wiring import In, Out
from amaranth.lib.memory import Memory, ReadPort, WritePort
from amaranth_soc import wishbone
from amaranth.utils import ceil_log2
import logging

from groovylight.geom import DisplayGeometry

from .common import Rgb666Layout, Hub75Stream, Hub75Ctrl, Hub75Data, Rgb888Layout


logger = logging.getLogger(__name__)


class SwapBuffer(wiring.Component):
    """A pair of BRAMs for holdling line data that are swapped between using an external signal.

    Contains one write port, and one read port. They are attached to separate BRAMs to allow for
    one to be read at the same time another is being updated.

    Parameters:
        depth: i32, depth of each memory buffer
        shape: ShapeLike, the underlying shape of the memory.

    Signals:
        write_port: WritePort.Signature(width, shape)
        read_port: ReadPort.Signature(width, shape)
        selector: In(1)

    """

    def __init__(self, shape: ShapeLike, depth: int, init=[]):
        self.mem_init = init
        super().__init__(
            {
                "selector": In(1),
                "read_port": Out(
                    ReadPort.Signature(addr_width=ceil_log2(depth), shape=shape)
                ),
                "write_port": Out(
                    WritePort.Signature(addr_width=ceil_log2(depth), shape=shape)
                ),
            }
        )
        self.data_shape = shape
        self.depth = depth

    def elaborate(self, platform: Platform) -> Module:
        m = Module()

        m.submodules.bram0 = self.bram0 = Memory(
            shape=self.data_shape, depth=self.depth, init=self.mem_init
        )
        m.submodules.bram1 = self.bram1 = Memory(
            shape=self.data_shape, depth=self.depth, init=self.mem_init
        )

        rd0 = self.bram0.read_port()
        wr0 = self.bram0.write_port()
        rd1 = self.bram1.read_port()
        wr1 = self.bram1.write_port()

        m.d.comb += [
            # wr addres
            wr0.addr.eq(self.write_port.addr),
            wr1.addr.eq(self.write_port.addr),
            # write enables are based on selector
            wr0.en.eq(~self.selector),
            wr1.en.eq(self.selector),
            # connect write data. This is FINE because
            # there is one driver (the external writer)
            # and we en based on selector so the other one isn't active
            wr0.data.eq(self.write_port.data),
            wr1.data.eq(self.write_port.data),
            # connect rd address lines
            rd0.addr.eq(self.read_port.addr),
            rd1.addr.eq(self.read_port.addr),
            rd0.en.eq(~self.selector),
            rd1.en.eq(self.selector),
            # we do this because the read_data lines are driven, this prevents
            # double-driver situations even though we en using selector above
            self.read_port.data.eq(Mux(self.selector, rd1.data, rd0.data)),
        ]

        return m


class DisplayClock(wiring.Component):
    """Generates the display clock automatically with the correct delay
    when start is asserted. Stops when `done` is strobed.

    Can either be used as a /2 clock divider or a /4 with the `double_fetch` parameter.
    The startup delay is chosen based on the mode or can be set with `startup_delay`
    """

    def __init__(self, *, double_fetch: bool = True, startup_delay=None, src_loc_at=0):
        self.double_fetch = double_fetch
        if startup_delay is None:
            self.startup_delay = 4 if double_fetch else 1  # FIXME: choose right values.
        else:
            logger.warn("Using custom startup delay. This is not tested")
            self.startup_delay = startup_delay

        super().__init__(
            {
                "start": In(1),
                "done": In(1),
                "display_clk": Out(1),
            },
            src_loc_at=src_loc_at,
        )

    def elaborate(self, platform: Platform) -> Module:
        m = Module()

        counter = Signal(range(max(self.startup_delay, 2) + 1))

        with m.FSM():
            with m.State("init"):
                m.d.sync += [counter.eq(0)]
                m.d.comb += self.display_clk.eq(0)
                with m.If(self.start == 1):
                    m.next = "warmup"
            with m.State("warmup"):
                m.d.sync += counter.eq(counter + 1)
                with m.If(counter == self.startup_delay - 1):
                    m.next = "run"
                    m.d.sync += counter.eq(0)

            with m.State("run"):
                with m.If(self.done == 1):
                    m.next = "init"
                if self.double_fetch:
                    m.d.comb += self.display_clk.eq(~counter[1])
                    m.d.sync += [
                        counter.eq(counter + 1),
                    ]
                else:
                    m.d.comb += (self.display_clk.eq(counter),)
                    m.d.sync += [
                        counter.eq(~counter),
                    ]

        return m


class DisplayStrober(wiring.Component):
    """Generates ~OE/blank strobes to expose the latched-in data."""

    def __init__(
        self,
        n_bits: int = 8,
        base_expose_time: int = None,
        *,
        periods=None,
        src_loc_at=0,
    ):
        self.n_bits = n_bits
        self.base_expose_time = base_expose_time
        if periods is None:
            # compute durations
            self.periods = Array([base_expose_time << x for x in range(n_bits)])
        else:
            assert len(periods) == n_bits, "must have one period for each bit"
            if base_expose_time is not None:
                logger.warn(
                    "Received both base_expose_time and predifined periods. Ignoring base_expose_time"
                )
            self.periods = Array(periods)
        logger.info("Strobe Clock Periods = %s", self.periods)
        super().__init__(
            {
                "start": In(1),
                "done": Out(1),
                "blank": Out(1),
                "bcm": In(ceil_log2(n_bits - 1)),
            },
            src_loc_at=src_loc_at,
        )

    def elaborate(self, platform: Platform) -> Module:
        m = Module()

        counter = Signal(32)
        _bcm = Signal(self.bcm)  # we save this value so that the coordinator can
        # start the next line while this is running.

        with m.FSM():
            with m.State("init"):
                m.d.sync += [counter.eq(0), self.done.eq(0), _bcm.eq(self.bcm)]
                m.d.comb += self.blank.eq(1)
                with m.If(self.start):
                    m.next = "run"
            with m.State("run"):
                m.d.comb += self.blank.eq(0)
                m.d.sync += counter.eq(counter + 1)
                with m.If(counter == self.periods[_bcm]):
                    m.next = "done"
            with m.State("done"):
                m.d.sync += self.done.eq(1)
                m.d.comb += self.blank.eq(1)
                m.next = "init"
        return m


class Hub75DataDriver(wiring.Component):
    def __init__(
        self,
        panel_length: int = 128,
        data_shape=Rgb888Layout,
        double_fetch=True,
        *,
        src_loc_at=0,
    ):
        self.panel_length = panel_length
        self.double_fetch = double_fetch
        # if we're using double_fetch then each bram cell is exactly one color,
        # if we're doing single fetch, it's 2 colors stacked.
        bram_shape = data_shape if double_fetch else data.ArrayLayout(data_shape, 2)
        addr_width = (
            ceil_log2(panel_length * 2) if double_fetch else ceil_log2(panel_length)
        )
        super().__init__(
            {
                "bcm_select": In(ceil_log2(data_shape["red"].shape.width)),
                "done": Out(1),
                "start": In(1),
                "bram": In(
                    ReadPort.Signature(
                        addr_width=addr_width,
                        shape=bram_shape,
                    )
                ),
                "data": Out(Hub75Data()),
            },
            src_loc_at=src_loc_at,
        )

    def elaborate(self, platform: Platform) -> Module:
        m = Module()

        counter = Signal(32)
        m.d.sync += counter.eq(counter + 1)

        pixnum = Signal(range(self.panel_length), init=self.panel_length - 1)
        if self.double_fetch:
            pixrow = Signal(1)
            m.d.comb += self.bram.addr.eq(Cat(pixrow, pixnum))
            ram_rgb_slice = self.bram.data.channel_slice(self.bcm_select)
        else:
            m.d.comb += self.bram.addr.eq(pixnum)
            ram_rgb_slice = Array(
                [
                    self.bram.data[0].channel_slice(self.bcm_select),
                    self.bram.data[1].channel_slice(self.bcm_select),
                ]
            )

        with m.FSM():
            with m.State("init"):
                m.d.sync += [
                    self.done.eq(0),
                    counter.eq(0),
                    pixnum.eq(pixnum.init),
                ]
                if self.double_fetch:
                    m.d.sync += pixrow.eq(0)
                with m.If(self.start == 1):
                    m.d.sync += self.bram.en.eq(1)
                    m.next = "prefetch" if self.double_fetch else "writerow"
            with m.State("prefetch"):
                # Allow the BRAM to settle after being enabled.
                m.d.sync += counter.eq(0)
                m.next = "writerow"

            with m.State("writerow"):
                if self.double_fetch:
                    c = counter[0:2]
                    with m.If(c == 0b0):
                        m.d.sync += [
                            self.data.rgb0.eq(ram_rgb_slice),
                            pixrow.eq(1),
                        ]
                    with m.If(c == 0b01):
                        pass
                    with m.If(c == 0b10):
                        m.d.sync += self.data.rgb1.eq(ram_rgb_slice)
                        m.d.sync += [
                            pixnum.eq(pixnum - 1),
                            pixrow.eq(0),
                        ]
                        with m.If(pixnum == 0):
                            m.next = "done"

                    with m.If(c == 0b11):
                        m.d.sync += counter.eq(0)
                else:
                    with m.If(counter[0] == 0):
                        m.d.sync += [
                            self.data.rgb0.eq(ram_rgb_slice[0]),
                            self.data.rgb1.eq(ram_rgb_slice[1]),
                            pixnum.eq(pixnum - 1),
                        ]
                    with m.If(counter[0] == 1):
                        m.d.sync += [
                            counter.eq(0),
                        ]
                        with m.If(pixnum == 0):
                            m.next = "done"
            with m.State("done"):
                m.d.sync += [self.done.eq(1), self.bram.en.eq(0)]
                m.next = "init"

        return m


class Hub75Coordinator(wiring.Component):
    """A shared-control hub75 driver"""

    def __init__(self, geom: DisplayGeometry, *, double_fetch=True):
        self.geom = geom
        self.double_fetch = double_fetch
        super().__init__(
            {
                "hub75": Out(Hub75Ctrl(self.geom.n_strings)),
                # TODO: fetching routine? maybe it's passed through.
            }
        )

    def elaborate(self, platform: Platform) -> Module:
        m = Module()
        # swapline is which buffer we are using vs sending out.
        swapline = Signal(1)

        # for each string, spawn a swapbuffer + stringdriver and connect.
        # don't worry about fetching for now.
        self.strings = strings = []
        self.buffers = bufs = []
        donearr = []
        startStrings = Signal(1)
        stringsDone = Signal(1)
        bram_shape = Rgb888Layout if self.double_fetch else data.ArrayLayout(Rgb888Layout, 2)

        for idx, string in enumerate(self.geom.strings):
            mdepth = string.dimensions.length
            if self.double_fetch:
                mdepth = mdepth * 2
            sb = SwapBuffer(depth=mdepth, shape=bram_shape)
            bufs.append(sb)
            stringdriver = Hub75DataDriver(
                string.dimensions.length,
                data_shape=Rgb888Layout,
                double_fetch=self.double_fetch,
            )
            strings.append(stringdriver)
            wiring.connect(m, sb.read_port, stringdriver.bram)
            # wiring.connect(m, self.hub75.data[idx], stringdriver.data.flip())
            m.d.comb += [
                self.hub75.data[idx].rgb0.eq(stringdriver.data.rgb0),
                self.hub75.data[idx].rgb1.eq(stringdriver.data.rgb1),
                stringdriver.start.eq(startStrings),
                sb.selector.eq(swapline),
            ]
            m.submodules += [sb, stringdriver]
            donearr.append(stringdriver.done)

        # combine the done signals into one signal with AND-reduction
        m.d.comb += stringsDone.eq(Cat(*donearr).all())

        self.addr = Signal(5)
        # handle the fetch side.
        # right now we're just going to use a basicFetcher
        # TODO: support SDRAM framebuffer

        with m.FSM():
            with m.State("init"):
                # go to preload.
                pass

            with m.State("preload"):
                pass

            with m.State("run"):
                pass

            with m.State("swap"):
                pass

        return m


# fetch line
